@@ -2,33 +2,35 @@
 """
 BLE Client for BlueZ on Linux
 """
-import inspect
-import logging
 import asyncio
+import logging
 import os
+import sys
 import warnings
-from typing import Callable, Optional, Union
+from typing import Callable, Dict, Optional, Set, Union, cast
 from uuid import UUID
 
-import async_timeout
+if sys.version_info < (3, 11):
+    from async_timeout import timeout as async_timeout
+else:
+    from asyncio import timeout as async_timeout
 
 from dbus_fast.aio import MessageBus
-from dbus_fast.constants import BusType, ErrorType
+from dbus_fast.constants import BusType, ErrorType, MessageType
 from dbus_fast.message import Message
 from dbus_fast.signature import Variant
 
-from bleak.backends.bluezdbus import defs
-from bleak.backends.bluezdbus.characteristic import BleakGATTCharacteristicBlueZDBus
-from bleak.backends.bluezdbus.manager import get_global_bluez_manager
-from bleak.backends.bluezdbus.scanner import BleakScannerBlueZDBus
-from bleak.backends.bluezdbus.utils import (
-    assert_reply,
-    extract_service_handle_from_path,
-)
-from bleak.backends.client import BaseBleakClient
-from bleak.backends.device import BLEDevice
-from bleak.backends.service import BleakGATTServiceCollection
-from bleak.exc import BleakDBusError, BleakError
+from ... import BleakScanner
+from ...exc import BleakDBusError, BleakError, BleakDeviceNotFoundError
+from ..characteristic import BleakGATTCharacteristic
+from ..client import BaseBleakClient, NotifyCallback
+from ..device import BLEDevice
+from ..service import BleakGATTServiceCollection
+from . import defs
+from .characteristic import BleakGATTCharacteristicBlueZDBus
+from .manager import get_global_bluez_manager
+from .scanner import BleakScannerBlueZDBus
+from .utils import assert_reply, get_dbus_authenticator
 from .version import BlueZFeatures
 
 logger = logging.getLogger(__name__)
@@ -41,6 +43,7 @@ class BleakClientBlueZDBus(BaseBleakClient):
 
     Args:
         address_or_ble_device (`BLEDevice` or str): The Bluetooth address of the BLE peripheral to connect to or the `BLEDevice` object representing it.
+        services: Optional list of service UUIDs that will be used.
 
     Keyword Args:
         timeout (float): Timeout for required ``BleakScanner.find_device_by_address`` call. Defaults to 10.0.
@@ -50,7 +53,12 @@ class BleakClientBlueZDBus(BaseBleakClient):
         adapter (str): Bluetooth adapter to use for discovery.
     """
 
-    def __init__(self, address_or_ble_device: Union[BLEDevice, str], **kwargs):
+    def __init__(
+        self,
+        address_or_ble_device: Union[BLEDevice, str],
+        services: Optional[Set[str]] = None,
+        **kwargs,
+    ):
         super(BleakClientBlueZDBus, self).__init__(address_or_ble_device, **kwargs)
         # kwarg "device" is for backwards compatibility
         self._adapter: Optional[str] = kwargs.get("adapter", kwargs.get("device"))
@@ -63,6 +71,8 @@ class BleakClientBlueZDBus(BaseBleakClient):
             self._device_path = None
             self._device_info = None
 
+        self._requested_services = services
+
         # D-Bus message bus
         self._bus: Optional[MessageBus] = None
         # tracks device watcher subscription
@@ -73,6 +83,8 @@ class BleakClientBlueZDBus(BaseBleakClient):
         self._disconnecting_event: Optional[asyncio.Event] = None
         # used to ensure device gets disconnected if event loop crashes
         self._disconnect_monitor_event: Optional[asyncio.Event] = None
+        # map of characteristic D-Bus object path to notification callback
+        self._notification_callbacks: Dict[str, NotifyCallback] = {}
 
         # used to override mtu_size property
         self._mtu_size: Optional[int] = None
@@ -106,129 +118,186 @@ class BleakClientBlueZDBus(BaseBleakClient):
         # Find the desired device before trying to connect.
         timeout = kwargs.get("timeout", self._timeout)
         if self._device_path is None:
-            device = await BleakScannerBlueZDBus.find_device_by_address(
-                self.address, timeout=timeout, adapter=self._adapter
+            device = await BleakScanner.find_device_by_address(
+                self.address,
+                timeout=timeout,
+                adapter=self._adapter,
+                backend=BleakScannerBlueZDBus,
             )
 
             if device:
                 self._device_info = device.details.get("props")
                 self._device_path = device.details["path"]
             else:
-                raise BleakError(
-                    "Device with address {0} was not found.".format(self.address)
+                raise BleakDeviceNotFoundError(
+                    self.address, f"Device with address {self.address} was not found."
                 )
 
         manager = await get_global_bluez_manager()
 
-        # Each BLE connection session needs a new D-Bus connection to avoid a
-        # BlueZ quirk where notifications are automatically enabled on reconnect.
-        self._bus = await MessageBus(
-            bus_type=BusType.SYSTEM, negotiate_unix_fd=True
-        ).connect()
+        async with async_timeout(timeout):
+            while True:
+                # Each BLE connection session needs a new D-Bus connection to avoid a
+                # BlueZ quirk where notifications are automatically enabled on reconnect.
+                self._bus = await MessageBus(
+                    bus_type=BusType.SYSTEM,
+                    negotiate_unix_fd=True,
+                    auth=get_dbus_authenticator(),
+                ).connect()
 
-        def on_connected_changed(connected: bool) -> None:
-            if not connected:
-                logger.debug(f"Device disconnected ({self._device_path})")
+                def on_connected_changed(connected: bool) -> None:
+                    if not connected:
+                        logger.debug(f"Device disconnected ({self._device_path})")
 
-                self._is_connected = False
+                        self._is_connected = False
 
-                if self._disconnect_monitor_event:
-                    self._disconnect_monitor_event.set()
-                    self._disconnect_monitor_event = None
+                        if self._disconnect_monitor_event:
+                            self._disconnect_monitor_event.set()
+                            self._disconnect_monitor_event = None
 
-                self._cleanup_all()
-                if self._disconnected_callback is not None:
-                    self._disconnected_callback(self)
-                disconnecting_event = self._disconnecting_event
-                if disconnecting_event:
-                    disconnecting_event.set()
+                        self._cleanup_all()
+                        if self._disconnected_callback is not None:
+                            self._disconnected_callback()
+                        disconnecting_event = self._disconnecting_event
+                        if disconnecting_event:
+                            disconnecting_event.set()
 
-        def on_value_changed(char_path: str, value: bytes) -> None:
-            if char_path in self._notification_callbacks:
-                handle = extract_service_handle_from_path(char_path)
-                self._notification_callbacks[char_path](handle, bytearray(value))
+                def on_value_changed(char_path: str, value: bytes) -> None:
+                    callback = self._notification_callbacks.get(char_path)
 
-        watcher = manager.add_device_watcher(
-            self._device_path, on_connected_changed, on_value_changed
-        )
-        self._remove_device_watcher = lambda: manager.remove_device_watcher(watcher)
+                    if callback:
+                        callback(bytearray(value))
 
-        try:
-            try:
-                #
-                # The BlueZ backend does not disconnect devices when the
-                # application closes or crashes. This can cause problems
-                # when trying to reconnect to the same device. To work
-                # around this, we check if the device is already connected.
-                #
-                # For additional details see https://github.com/bluez/bluez/issues/89
-                #
-                if not manager.is_connected(self._device_path):
-                    async with async_timeout.timeout(timeout):
-                        reply = await self._bus.call(
-                            Message(
-                                destination=defs.BLUEZ_SERVICE,
-                                interface=defs.DEVICE_INTERFACE,
-                                path=self._device_path,
-                                member="Connect",
-                            )
-                        )
-                    assert_reply(reply)
-
-                self._is_connected = True
-
-                # Create a task that runs until the device is disconnected.
-                self._disconnect_monitor_event = asyncio.Event()
-                asyncio.ensure_future(
-                    self._disconnect_monitor(
-                        self._bus, self._device_path, self._disconnect_monitor_event
-                    )
+                watcher = manager.add_device_watcher(
+                    self._device_path, on_connected_changed, on_value_changed
+                )
+                self._remove_device_watcher = lambda: manager.remove_device_watcher(
+                    watcher
                 )
 
-                #
-                # We will try to use the cache if it exists and `dangerous_use_bleak_cache`
-                # is True.
-                #
-                await self.get_services(
-                    dangerous_use_bleak_cache=dangerous_use_bleak_cache
-                )
+                self._disconnect_monitor_event = (
+                    local_disconnect_monitor_event
+                ) = asyncio.Event()
 
-                return True
-            except BaseException:
-                # Calling Disconnect cancels any pending connect request. Also,
-                # if connection was successful but get_services() raises (e.g.
-                # because task was cancelled), the we still need to disconnect
-                # before passing on the exception.
-                if self._bus:
-                    # If disconnected callback already fired, this will be a no-op
-                    # since self._bus will be None and the _cleanup_all call will
-                    # have already disconnected.
+                try:
                     try:
-                        reply = await self._bus.call(
-                            Message(
-                                destination=defs.BLUEZ_SERVICE,
-                                interface=defs.DEVICE_INTERFACE,
-                                path=self._device_path,
-                                member="Disconnect",
+                        #
+                        # The BlueZ backend does not disconnect devices when the
+                        # application closes or crashes. This can cause problems
+                        # when trying to reconnect to the same device. To work
+                        # around this, we check if the device is already connected.
+                        #
+                        # For additional details see https://github.com/bluez/bluez/issues/89
+                        #
+                        if manager.is_connected(self._device_path):
+                            logger.debug(
+                                'skipping calling "Connect" since %s is already connected',
+                                self._device_path,
+                            )
+                        else:
+                            logger.debug(
+                                "Connecting to BlueZ path %s", self._device_path
+                            )
+                            reply = await self._bus.call(
+                                Message(
+                                    destination=defs.BLUEZ_SERVICE,
+                                    interface=defs.DEVICE_INTERFACE,
+                                    path=self._device_path,
+                                    member="Connect",
+                                )
+                            )
+
+                            assert reply is not None
+
+                            if reply.message_type == MessageType.ERROR:
+                                # This error is often caused by RF interference
+                                # from other Bluetooth or Wi-Fi devices. In many
+                                # cases, retrying will connect successfully.
+                                # Note: this error was added in BlueZ 6.62.
+                                if (
+                                    reply.error_name == "org.bluez.Error.Failed"
+                                    and reply.body
+                                    and reply.body[0] == "le-connection-abort-by-local"
+                                ):
+                                    logger.debug(
+                                        "retry due to le-connection-abort-by-local"
+                                    )
+
+                                    # When this error occurs, BlueZ actually
+                                    # connected so we get "Connected" property changes
+                                    # that we need to wait for before attempting
+                                    # to connect again.
+                                    await local_disconnect_monitor_event.wait()
+
+                                    # Jump way back to the `while True:`` to retry.
+                                    continue
+
+                                if reply.error_name == ErrorType.UNKNOWN_OBJECT.value:
+                                    raise BleakDeviceNotFoundError(
+                                        self.address,
+                                        f"Device with address {self.address} was not found. It may have been removed from BlueZ when scanning stopped.",
+                                    )
+
+                            assert_reply(reply)
+
+                        self._is_connected = True
+
+                        # Create a task that runs until the device is disconnected.
+                        asyncio.ensure_future(
+                            self._disconnect_monitor(
+                                self._bus,
+                                self._device_path,
+                                local_disconnect_monitor_event,
                             )
                         )
-                        try:
-                            assert_reply(reply)
-                        except BleakDBusError as e:
-                            # if the object no longer exists, then we know we
-                            # are disconnected for sure, so don't need to log a
-                            # warning about it
-                            if e.dbus_error != ErrorType.UNKNOWN_OBJECT.value:
-                                raise
-                    except Exception as e:
-                        logger.warning(
-                            f"Failed to cancel connection ({self._device_path}): {e}"
+
+                        #
+                        # We will try to use the cache if it exists and `dangerous_use_bleak_cache`
+                        # is True.
+                        #
+                        await self.get_services(
+                            dangerous_use_bleak_cache=dangerous_use_bleak_cache
                         )
 
-                raise
-        except BaseException:
-            self._cleanup_all()
-            raise
+                        return True
+                    except BaseException:
+                        # Calling Disconnect cancels any pending connect request. Also,
+                        # if connection was successful but get_services() raises (e.g.
+                        # because task was cancelled), the we still need to disconnect
+                        # before passing on the exception.
+                        if self._bus:
+                            # If disconnected callback already fired, this will be a no-op
+                            # since self._bus will be None and the _cleanup_all call will
+                            # have already disconnected.
+                            try:
+                                reply = await self._bus.call(
+                                    Message(
+                                        destination=defs.BLUEZ_SERVICE,
+                                        interface=defs.DEVICE_INTERFACE,
+                                        path=self._device_path,
+                                        member="Disconnect",
+                                    )
+                                )
+                                try:
+                                    assert_reply(reply)
+                                except BleakDBusError as e:
+                                    # if the object no longer exists, then we know we
+                                    # are disconnected for sure, so don't need to log a
+                                    # warning about it
+                                    if e.dbus_error != ErrorType.UNKNOWN_OBJECT.value:
+                                        raise
+                            except Exception as e:
+                                logger.warning(
+                                    f"Failed to cancel connection ({self._device_path}): {e}"
+                                )
+
+                        raise
+                except BaseException:
+                    # this effectively cancels the disconnect monitor in case the event
+                    # was not triggered by a D-Bus callback
+                    local_disconnect_monitor_event.set()
+                    self._cleanup_all()
+                    raise
 
     @staticmethod
     async def _disconnect_monitor(
@@ -290,8 +359,7 @@ class BleakClientBlueZDBus(BaseBleakClient):
             self._bus = None
 
             # Reset all stored services.
-            self.services = BleakGATTServiceCollection()
-            self._services_resolved = False
+            self.services = None
 
     async def disconnect(self) -> bool:
         """Disconnect from the specified GATT server.
@@ -315,7 +383,7 @@ class BleakClientBlueZDBus(BaseBleakClient):
         if self._disconnecting_event:
             # another call to disconnect() is already in progress
             logger.debug(f"already in progress ({self._device_path})")
-            async with async_timeout.timeout(10):
+            async with async_timeout(10):
                 await self._disconnecting_event.wait()
         elif self.is_connected:
             self._disconnecting_event = asyncio.Event()
@@ -330,7 +398,7 @@ class BleakClientBlueZDBus(BaseBleakClient):
                     )
                 )
                 assert_reply(reply)
-                async with async_timeout.timeout(10):
+                async with async_timeout(10):
                     await self._disconnecting_event.wait()
             finally:
                 self._disconnecting_event = None
@@ -413,10 +481,46 @@ class BleakClientBlueZDBus(BaseBleakClient):
             Boolean regarding success of unpairing.
 
         """
-        warnings.warn(
-            "Unpairing is seemingly unavailable in the BlueZ DBus API at the moment."
+        adapter_path = await self._get_adapter_path()
+        device_path = await self._get_device_path()
+        manager = await get_global_bluez_manager()
+
+        logger.debug(
+            "Removing BlueZ device path %s from adapter path %s",
+            device_path,
+            adapter_path,
         )
-        return False
+
+        # If this client object wants to connect again, BlueZ needs the device
+        # to follow Discovery process again - so reset the local connection
+        # state.
+        #
+        # (This is true even if the request to RemoveDevice fails,
+        # so clear it before.)
+        self._device_path = None
+        self._device_info = None
+        self._is_connected = False
+
+        try:
+            reply = await manager._bus.call(
+                Message(
+                    destination=defs.BLUEZ_SERVICE,
+                    path=adapter_path,
+                    interface=defs.ADAPTER_INTERFACE,
+                    member="RemoveDevice",
+                    signature="o",
+                    body=[device_path],
+                )
+            )
+            assert_reply(reply)
+        except BleakDBusError as e:
+            if e.dbus_error == "org.bluez.Error.DoesNotExist":
+                raise BleakDeviceNotFoundError(
+                    self.address, f"Device with address {self.address} was not found."
+                ) from e
+            raise
+
+        return True
 
     @property
     def is_connected(self) -> bool:
@@ -473,6 +577,37 @@ class BleakClientBlueZDBus(BaseBleakClient):
         os.close(reply.unix_fds[0])
         self._mtu_size = reply.body[1]
 
+    async def _get_adapter_path(self) -> str:
+        """Private coroutine to return the BlueZ path to the adapter this client is assigned to.
+
+        Can be called even if no connection has been established yet.
+        """
+        if self._device_info:
+            # If we have a BlueZ DBus object with _device_info, use what it tell us
+            return self._device_info["Adapter"]
+        if self._adapter:
+            # If the adapter name was set in the constructor, convert to a BlueZ path
+            return f"/org/bluez/{self._adapter}"
+
+        # Fall back to the system's default Bluetooth adapter
+        manager = await get_global_bluez_manager()
+        return manager.get_default_adapter()
+
+    async def _get_device_path(self) -> str:
+        """Private coroutine to return the BlueZ path to the device address this client is assigned to.
+
+        Unlike the _device_path property, this function can be called even if discovery process has not
+        started and/or connection has not been established yet.
+        """
+        if self._device_path:
+            # If we have a BlueZ DBus object, return its device path
+            return self._device_path
+
+        # Otherwise, build a new path using the adapter path and the BLE address
+        adapter_path = await self._get_adapter_path()
+        bluez_address = self.address.upper().replace(":", "_")
+        return f"{adapter_path}/dev_{bluez_address}"
+
     @property
     def mtu_size(self) -> int:
         """Get ATT MTU size for active connection"""
@@ -501,15 +636,14 @@ class BleakClientBlueZDBus(BaseBleakClient):
         if not self.is_connected:
             raise BleakError("Not connected")
 
-        if self._services_resolved:
+        if self.services is not None:
             return self.services
 
         manager = await get_global_bluez_manager()
 
         self.services = await manager.get_services(
-            self._device_path, dangerous_use_bleak_cache
+            self._device_path, dangerous_use_bleak_cache, self._requested_services
         )
-        self._services_resolved = True
 
         return self.services
 
@@ -780,65 +914,18 @@ class BleakClientBlueZDBus(BaseBleakClient):
 
     async def start_notify(
         self,
-        char_specifier: Union[BleakGATTCharacteristicBlueZDBus, int, str, UUID],
-        callback: Callable[[int, bytearray], None],
+        characteristic: BleakGATTCharacteristic,
+        callback: NotifyCallback,
         **kwargs,
     ) -> None:
-        """Activate notifications/indications on a characteristic.
-
-        Callbacks must accept two inputs. The first will be a integer handle of the characteristic generating the
-        data and the second will be a ``bytearray`` containing the data sent from the connected server.
-
-        .. code-block:: python
-
-            def callback(sender: int, data: bytearray):
-                print(f"{sender}: {data}")
-            client.start_notify(char_uuid, callback)
-
-        Args:
-            char_specifier (BleakGATTCharacteristicBlueZDBus, int, str or UUID): The characteristic to activate
-                notifications/indications on a characteristic, specified by either integer handle,
-                UUID or directly by the BleakGATTCharacteristicBlueZDBus object representing it.
-            callback (function): The function to be called on notification.
         """
-        if not self.is_connected:
-            raise BleakError("Not connected")
+        Activate notifications/indications on a characteristic.
+        """
+        characteristic = cast(BleakGATTCharacteristicBlueZDBus, characteristic)
 
-        if inspect.iscoroutinefunction(callback):
+        self._notification_callbacks[characteristic.path] = callback
 
-            def bleak_callback(s, d):
-                asyncio.ensure_future(callback(s, d))
-
-        else:
-            bleak_callback = callback
-
-        if not isinstance(char_specifier, BleakGATTCharacteristicBlueZDBus):
-            characteristic = self.services.get_characteristic(char_specifier)
-        else:
-            characteristic = char_specifier
-
-        if not characteristic:
-            # Special handling for BlueZ >= 5.48, where Battery Service (0000180f-0000-1000-8000-00805f9b34fb:)
-            # has been moved to interface org.bluez.Battery1 instead of as a regular service.
-            # The org.bluez.Battery1 on the other hand does not provide a notification method, so here we cannot
-            # provide this functionality...
-            # See https://kernel.googlesource.com/pub/scm/bluetooth/bluez/+/refs/tags/5.48/doc/battery-api.txt
-            if str(char_specifier) == "00002a19-0000-1000-8000-00805f9b34fb" and (
-                BlueZFeatures.hides_battery_characteristic
-            ):
-                raise BleakError(
-                    "Notifications on Battery Level Char ({0}) is not "
-                    "possible in BlueZ >= 5.48. Use regular read instead.".format(
-                        char_specifier
-                    )
-                )
-            raise BleakError(
-                "Characteristic with UUID {0} could not be found!".format(
-                    char_specifier
-                )
-            )
-
-        self._notification_callbacks[characteristic.path] = bleak_callback
+        assert self._bus is not None
 
         reply = await self._bus.call(
             Message(
